@@ -1,4 +1,4 @@
-/*  -*- coding: utf-8 -*- */
+/*  -*- coding: utf-8; mode: c; mode: orgtbl -*- */
 /*
  * Mandos plugin runner - Run Mandos plugins
  *
@@ -38,7 +38,8 @@
 #include <sys/select.h>		/* fd_set, select(), FD_ZERO(),
 				   FD_SET(), FD_ISSET(), FD_CLR */
 #include <sys/wait.h>		/* wait(), waitpid(), WIFEXITED(),
-				   WEXITSTATUS() */
+				   WEXITSTATUS(), WTERMSIG(),
+				   WCOREDUMP() */
 #include <sys/stat.h>		/* struct stat, stat(), S_ISREG() */
 #include <iso646.h>		/* and, or, not */
 #include <dirent.h>		/* DIR, struct dirent, opendir(),
@@ -52,7 +53,8 @@
 				   close() */
 #include <fcntl.h>		/* fcntl(), F_GETFD, F_SETFD,
 				   FD_CLOEXEC */
-#include <string.h>		/* strsep, strlen(), asprintf() */
+#include <string.h>		/* strsep, strlen(), asprintf(),
+				   strsignal() */
 #include <errno.h>		/* errno */
 #include <argp.h>		/* struct argp_option, struct
 				   argp_state, struct argp,
@@ -108,13 +110,18 @@ static plugin *getplugin(char *name){
     }
   }
   /* Create a new plugin */
-  plugin *new_plugin = malloc(sizeof(plugin));
+  plugin *new_plugin = NULL;
+  do {
+    new_plugin = malloc(sizeof(plugin));
+  } while(new_plugin == NULL and errno == EINTR);
   if(new_plugin == NULL){
     return NULL;
   }
   char *copy_name = NULL;
   if(name != NULL){
-    copy_name = strdup(name);
+    do {
+      copy_name = strdup(name);
+    } while(copy_name == NULL and errno == EINTR);
     if(copy_name == NULL){
       free(new_plugin);
       return NULL;
@@ -126,7 +133,9 @@ static plugin *getplugin(char *name){
 			  .disabled = false,
 			  .next = plugin_list };
   
-  new_plugin->argv = malloc(sizeof(char *) * 2);
+  do {
+    new_plugin->argv = malloc(sizeof(char *) * 2);
+  } while(new_plugin->argv == NULL and errno == EINTR);
   if(new_plugin->argv == NULL){
     free(copy_name);
     free(new_plugin);
@@ -135,7 +144,9 @@ static plugin *getplugin(char *name){
   new_plugin->argv[0] = copy_name;
   new_plugin->argv[1] = NULL;
   
-  new_plugin->environ = malloc(sizeof(char *));
+  do {
+    new_plugin->environ = malloc(sizeof(char *));
+  } while(new_plugin->environ == NULL and errno == EINTR);
   if(new_plugin->environ == NULL){
     free(copy_name);
     free(new_plugin->argv);
@@ -153,14 +164,19 @@ static plugin *getplugin(char *name){
 static bool add_to_char_array(const char *new, char ***array,
 			      int *len){
   /* Resize the pointed-to array to hold one more pointer */
-  *array = realloc(*array, sizeof(char *)
-		   * (size_t) ((*len) + 2));
+  do {
+    *array = realloc(*array, sizeof(char *)
+		     * (size_t) ((*len) + 2));
+  } while(*array == NULL and errno == EINTR);
   /* Malloc check */
   if(*array == NULL){
     return false;
   }
   /* Make a copy of the new string */
-  char *copy = strdup(new);
+  char *copy;
+  do {
+    copy = strdup(new);
+  } while(copy == NULL and errno == EINTR);
   if(copy == NULL){
     return false;
   }
@@ -192,7 +208,10 @@ static bool add_environment(plugin *p, const char *def, bool replace){
     if(strncmp(*e, def, namelen + 1) == 0){
       /* It already exists */
       if(replace){
-	char *new = realloc(*e, strlen(def) + 1);
+	char *new;
+	do {
+	  new = realloc(*e, strlen(def) + 1);
+	} while(new == NULL and errno == EINTR);
 	if(new == NULL){
 	  return false;
 	}
@@ -208,16 +227,17 @@ static bool add_environment(plugin *p, const char *def, bool replace){
 /*
  * Based on the example in the GNU LibC manual chapter 13.13 "File
  * Descriptor Flags".
- * *Note File Descriptor Flags:(libc)Descriptor Flags.
+ | [[info:libc:Descriptor%20Flags][File Descriptor Flags]] |
  */
 static int set_cloexec_flag(int fd){
-  int ret = fcntl(fd, F_GETFD, 0);
+  int ret = (int)TEMP_FAILURE_RETRY(fcntl(fd, F_GETFD, 0));
   /* If reading the flags failed, return error indication now. */
   if(ret < 0){
     return ret;
   }
   /* Store modified flag word in the descriptor. */
-  return fcntl(fd, F_SETFD, ret | FD_CLOEXEC);
+  return (int)TEMP_FAILURE_RETRY(fcntl(fd, F_SETFD,
+				       ret | FD_CLOEXEC));
 }
 
 
@@ -315,7 +335,6 @@ int main(int argc, char *argv[]){
   fd_set rfds_all;
   int ret, maxfd = 0;
   ssize_t sret;
-  intmax_t tmpmax;
   uid_t uid = 65534;
   gid_t gid = 65534;
   bool debug = false;
@@ -380,8 +399,9 @@ int main(int argc, char *argv[]){
   
   error_t parse_opt(int key, char *arg, __attribute__((unused))
 		    struct argp_state *state){
-    char *tmp;
     switch(key){
+      char *tmp;
+      intmax_t tmpmax;
     case 'g': 			/* --global-options */
       if(arg != NULL){
 	char *plugin_option;
@@ -457,7 +477,7 @@ int main(int argc, char *argv[]){
       plugindir = strdup(arg);
       if(plugindir == NULL){
 	perror("strdup");
-      }      
+      }
       break;
     case 129:			/* --config-file */
       /* This is already done by parse_opt_config_file() */
@@ -527,7 +547,7 @@ int main(int argc, char *argv[]){
       if(argfile == NULL){
 	perror("strdup");
       }
-      break;      
+      break;
     case 130:			/* --userid */
     case 131:			/* --groupid */
     case 132:			/* --debug */
@@ -562,7 +582,7 @@ int main(int argc, char *argv[]){
     conffp = fopen(AFILE, "r");
   } else {
     conffp = fopen(argfile, "r");
-  }  
+  }
   if(conffp != NULL){
     char *org_line = NULL;
     char *p, *arg, *new_arg, *line;
@@ -612,9 +632,17 @@ int main(int argc, char *argv[]){
 	  goto fallback;
 	}
 	custom_argv[custom_argc-1] = new_arg;
-	custom_argv[custom_argc] = NULL;	
+	custom_argv[custom_argc] = NULL;
       }
     }
+    do {
+      ret = fclose(conffp);
+    } while(ret == EOF and errno == EINTR);
+    if(ret == EOF){
+      perror("fclose");
+      exitstatus = EXIT_FAILURE;
+      goto fallback;
+    }
     free(org_line);
   } else {
     /* Check for harmful errors and go to fallback. Other errors might
@@ -699,7 +727,9 @@ int main(int argc, char *argv[]){
   
   /* Read and execute any executable in the plugin directory*/
   while(true){
-    dirst = readdir(dir);
+    do {
+      dirst = readdir(dir);
+    } while(dirst == NULL and errno == EINTR);
     
     /* All directory entries have been processed */
     if(dirst == NULL){
@@ -759,16 +789,19 @@ int main(int argc, char *argv[]){
     
     char *filename;
     if(plugindir == NULL){
-      ret = asprintf(&filename, PDIR "/%s", dirst->d_name);
+      ret = (int)TEMP_FAILURE_RETRY(asprintf(&filename, PDIR "/%s",
+					     dirst->d_name));
     } else {
-      ret = asprintf(&filename, "%s/%s", plugindir, dirst->d_name);
+      ret = (int)TEMP_FAILURE_RETRY(asprintf(&filename, "%s/%s",
+					     plugindir,
+					     dirst->d_name));
     }
     if(ret < 0){
       perror("asprintf");
       continue;
     }
     
-    ret = stat(filename, &st);
+    ret = (int)TEMP_FAILURE_RETRY(stat(filename, &st));
     if(ret == -1){
       perror("stat");
       free(filename);
@@ -776,7 +809,8 @@ int main(int argc, char *argv[]){
     }
     
     /* Ignore non-executable files */
-    if(not S_ISREG(st.st_mode) or (access(filename, X_OK) != 0)){
+    if(not S_ISREG(st.st_mode)
+       or (TEMP_FAILURE_RETRY(access(filename, X_OK)) != 0)){
       if(debug){
 	fprintf(stderr, "Ignoring plugin dir entry \"%s\""
 		" with bad type or mode\n", filename);
@@ -828,7 +862,7 @@ int main(int argc, char *argv[]){
     }
     
     int pipefd[2];
-    ret = pipe(pipefd);
+    ret = (int)TEMP_FAILURE_RETRY(pipe(pipefd));
     if(ret == -1){
       perror("pipe");
       exitstatus = EXIT_FAILURE;
@@ -848,14 +882,19 @@ int main(int argc, char *argv[]){
       goto fallback;
     }
     /* Block SIGCHLD until process is safely in process list */
-    ret = sigprocmask(SIG_BLOCK, &sigchld_action.sa_mask, NULL);
+    ret = (int)TEMP_FAILURE_RETRY(sigprocmask(SIG_BLOCK,
+					      &sigchld_action.sa_mask,
+					      NULL));
     if(ret < 0){
       perror("sigprocmask");
       exitstatus = EXIT_FAILURE;
       goto fallback;
     }
     /* Starting a new process to be watched */
-    pid_t pid = fork();
+    pid_t pid;
+    do {
+      pid = fork();
+    } while(pid == -1 and errno == EINTR);
     if(pid == -1){
       perror("fork");
       exitstatus = EXIT_FAILURE;
@@ -899,12 +938,15 @@ int main(int argc, char *argv[]){
       /* no return */
     }
     /* Parent process */
-    close(pipefd[1]);		/* Close unused write end of pipe */
+    TEMP_FAILURE_RETRY(close(pipefd[1])); /* Close unused write end of
+					     pipe */
     free(filename);
     plugin *new_plugin = getplugin(dirst->d_name);
     if(new_plugin == NULL){
       perror("getplugin");
-      ret = sigprocmask(SIG_UNBLOCK, &sigchld_action.sa_mask, NULL);
+      ret = (int)(TEMP_FAILURE_RETRY
+		  (sigprocmask(SIG_UNBLOCK, &sigchld_action.sa_mask,
+			       NULL)));
       if(ret < 0){
         perror("sigprocmask");
       }
@@ -917,7 +959,9 @@ int main(int argc, char *argv[]){
     
     /* Unblock SIGCHLD so signal handler can be run if this process
        has already completed */
-    ret = sigprocmask(SIG_UNBLOCK, &sigchld_action.sa_mask, NULL);
+    ret = (int)TEMP_FAILURE_RETRY(sigprocmask(SIG_UNBLOCK,
+					      &sigchld_action.sa_mask,
+					      NULL));
     if(ret < 0){
       perror("sigprocmask");
       exitstatus = EXIT_FAILURE;
@@ -931,7 +975,7 @@ int main(int argc, char *argv[]){
     }
   }
   
-  closedir(dir);
+  TEMP_FAILURE_RETRY(closedir(dir));
   dir = NULL;
   free_plugin(getplugin(NULL));
   
@@ -950,7 +994,7 @@ int main(int argc, char *argv[]){
   while(plugin_list){
     fd_set rfds = rfds_all;
     int select_ret = select(maxfd+1, &rfds, NULL, NULL, NULL);
-    if(select_ret == -1){
+    if(select_ret == -1 and errno != EINTR){
       perror("select");
       exitstatus = EXIT_FAILURE;
       goto fallback;
@@ -973,9 +1017,10 @@ int main(int argc, char *argv[]){
 		      WEXITSTATUS(proc->status));
 	    } else if(WIFSIGNALED(proc->status)){
 	      fprintf(stderr, "Plugin %s [%" PRIdMAX "] killed by"
-		      " signal %d\n", proc->name,
+		      " signal %d: %s\n", proc->name,
 		      (intmax_t) (proc->pid),
-		      WTERMSIG(proc->status));
+		      WTERMSIG(proc->status),
+		      strsignal(WTERMSIG(proc->status)));
 	    } else if(WCOREDUMP(proc->status)){
 	      fprintf(stderr, "Plugin %s [%" PRIdMAX "] dumped"
 		      " core\n", proc->name, (intmax_t) (proc->pid));
@@ -986,7 +1031,10 @@ int main(int argc, char *argv[]){
 	  FD_CLR(proc->fd, &rfds_all);
 	  
 	  /* Block signal while modifying process_list */
-	  ret = sigprocmask(SIG_BLOCK, &sigchld_action.sa_mask, NULL);
+	  ret = (int)TEMP_FAILURE_RETRY(sigprocmask
+					(SIG_BLOCK,
+					 &sigchld_action.sa_mask,
+					 NULL));
 	  if(ret < 0){
 	    perror("sigprocmask");
 	    exitstatus = EXIT_FAILURE;
@@ -998,8 +1046,9 @@ int main(int argc, char *argv[]){
 	  proc = next_plugin;
 	  
 	  /* We are done modifying process list, so unblock signal */
-	  ret = sigprocmask(SIG_UNBLOCK, &sigchld_action.sa_mask,
-			    NULL);
+	  ret = (int)(TEMP_FAILURE_RETRY
+		      (sigprocmask(SIG_UNBLOCK,
+				   &sigchld_action.sa_mask, NULL)));
 	  if(ret < 0){
 	    perror("sigprocmask");
 	    exitstatus = EXIT_FAILURE;
@@ -1042,8 +1091,10 @@ int main(int argc, char *argv[]){
 	proc->buffer_size += BUFFER_SIZE;
       }
       /* Read from the process */
-      sret = read(proc->fd, proc->buffer + proc->buffer_length,
-		  BUFFER_SIZE);
+      sret = TEMP_FAILURE_RETRY(read(proc->fd,
+				     proc->buffer
+				     + proc->buffer_length,
+				     BUFFER_SIZE));
       if(sret < 0){
 	/* Read error from this process; ignore the error */
 	proc = proc->next;
@@ -1111,7 +1162,7 @@ int main(int argc, char *argv[]){
   }
   
   /* Wait for any remaining child processes to terminate */
-  do{
+  do {
     ret = wait(NULL);
   } while(ret >= 0);
   if(errno != ECHILD){
